/* 
 *  vobdec.cpp: implementation of the descrambling stuff.
 *
 *	based on vobdec.c source (authour anonymous?), modified by hrm.
 *
 *  This file is part of FlasKMPEG, a free MPEG to MPEG/AVI converter
 *	
 *  FlasKMPEG is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  FlasKMPEG is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "inputStream.h"

static unsigned char	_reverse[0x100];
static unsigned char	_table[0x100] = {
	0x33, 0x73, 0x3B, 0x26, 0x63, 0x23, 0x6B, 0x76, 0x3E, 0x7E, 0x36, 0x2B, 0x6E, 0x2E, 0x66, 0x7B,
	0xD3, 0x93, 0xDB, 0x06, 0x43, 0x03, 0x4B, 0x96, 0xDE, 0x9E, 0xD6, 0x0B, 0x4E, 0x0E, 0x46, 0x9B,
	0x57, 0x17, 0x5F, 0x82, 0xC7, 0x87, 0xCF, 0x12, 0x5A, 0x1A, 0x52, 0x8F, 0xCA, 0x8A, 0xC2, 0x1F,
	0xD9, 0x99, 0xD1, 0x00, 0x49, 0x09, 0x41, 0x90, 0xD8, 0x98, 0xD0, 0x01, 0x48, 0x08, 0x40, 0x91,
	0x3D, 0x7D, 0x35, 0x24, 0x6D, 0x2D, 0x65, 0x74, 0x3C, 0x7C, 0x34, 0x25, 0x6C, 0x2C, 0x64, 0x75,
	0xDD, 0x9D, 0xD5, 0x04, 0x4D, 0x0D, 0x45, 0x94, 0xDC, 0x9C, 0xD4, 0x05, 0x4C, 0x0C, 0x44, 0x95,
	0x59, 0x19, 0x51, 0x80, 0xC9, 0x89, 0xC1, 0x10, 0x58, 0x18, 0x50, 0x81, 0xC8, 0x88, 0xC0, 0x11,
	0xD7, 0x97, 0xDF, 0x02, 0x47, 0x07, 0x4F, 0x92, 0xDA, 0x9A, 0xD2, 0x0F, 0x4A, 0x0A, 0x42, 0x9F,
	0x53, 0x13, 0x5B, 0x86, 0xC3, 0x83, 0xCB, 0x16, 0x5E, 0x1E, 0x56, 0x8B, 0xCE, 0x8E, 0xC6, 0x1B,
	0xB3, 0xF3, 0xBB, 0xA6, 0xE3, 0xA3, 0xEB, 0xF6, 0xBE, 0xFE, 0xB6, 0xAB, 0xEE, 0xAE, 0xE6, 0xFB,
	0x37, 0x77, 0x3F, 0x22, 0x67, 0x27, 0x6F, 0x72, 0x3A, 0x7A, 0x32, 0x2F, 0x6A, 0x2A, 0x62, 0x7F,
	0xB9, 0xF9, 0xB1, 0xA0, 0xE9, 0xA9, 0xE1, 0xF0, 0xB8, 0xF8, 0xB0, 0xA1, 0xE8, 0xA8, 0xE0, 0xF1,
	0x5D, 0x1D, 0x55, 0x84, 0xCD, 0x8D, 0xC5, 0x14, 0x5C, 0x1C, 0x54, 0x85, 0xCC, 0x8C, 0xC4, 0x15,
	0xBD, 0xFD, 0xB5, 0xA4, 0xED, 0xAD, 0xE5, 0xF4, 0xBC, 0xFC, 0xB4, 0xA5, 0xEC, 0xAC, 0xE4, 0xF5,
	0x39, 0x79, 0x31, 0x20, 0x69, 0x29, 0x61, 0x70, 0x38, 0x78, 0x30, 0x21, 0x68, 0x28, 0x60, 0x71,
	0xB7, 0xF7, 0xBF, 0xA2, 0xE7, 0xA7, 0xEF, 0xF2, 0xBA, 0xFA, 0xB2, 0xAF, 0xEA, 0xAA, 0xE2, 0xFF,
};

//----------------------------------------------------------------------------------------------------------

// global (yuck!)
VobRunTime vrt;

/* begin of stuff from vobdec.c */
inline void _ClockLfsr0Forward(int &lfsr0)
{
	int	temp;

	temp = (lfsr0 << 3) | (lfsr0 >> 14);
	lfsr0 = (lfsr0 >> 8) | ((((((temp << 3) ^ temp) << 3) ^ temp ^ lfsr0) & 0xFF) << 9);
}

inline void _ClockLfsr1Forward(int &lfsr1)
{
	lfsr1 = (lfsr1 >> 8) | ((((((((lfsr1 >> 8) ^ lfsr1) >> 1) ^ lfsr1) >> 3) ^ lfsr1) & 0xFF) << 17);
}

void _Salt(int &lfsr0, int &lfsr1, const unsigned char salt[5])
{
	lfsr0 ^= (_reverse[salt[0]] << 9) | _reverse[salt[1]];
	lfsr1 ^= ((_reverse[salt[2]] & 0xE0) << 17) | ((_reverse[salt[2]] & 0x1F) << 16) | (_reverse[salt[3]] << 8) | _reverse[salt[4]];
}

static void _Decrypt(unsigned char *buffer, int lfsr0, int lfsr1)
{
	int	loop0, carry;

	_Salt(lfsr0, lfsr1, buffer + 0x54);
	for(loop0 = carry = 0, buffer += 0x80; loop0 != 0x800 - 0x80; loop0++, buffer++) {
		_ClockLfsr0Forward(lfsr0);
		_ClockLfsr1Forward(lfsr1);
		carry += ((lfsr0 >> 9) ^ 0xFF) + (lfsr1 >> 17);
		*buffer = (unsigned char) (_table[*buffer] ^ carry);
		carry >>= 8;
	}
}
/* end of stuff from vobdec.c */

static void _BufDecrypt(unsigned char* buffer, int size, int lfsr0, int lfsr1)
{
	while (size >= 0x800) {
		if (buffer[0x14] & 0x10) {
			_Decrypt(buffer, lfsr0, lfsr1);
			buffer[0x14] &= ~0x10;
		}
		size -= 0x800;
		buffer += 0x800;
	}
}

// return values: 0: not scrambled (at least not first num_blocks), 1: scrambled, 
// 2: error occurred, see message.
// stores a ptr to a mesg in msgptr (only if ret val is 2)
// stores a hash over num_blocks in hash (used to ID the file)
int ExamineVob(const char *filename, int num_blocks, unsigned long& hash, const char* &msgptr)
{
	unsigned char	buffer[0x800]; 
	int				block, result;
	FILE			*file;
	static char		msg[1024];

	sprintf(msg, "No error");
	msgptr = msg;
	hash = 0;
	result = 0; // not scrambled, until proven otherwise...
	if((file = fopen(filename, "rb")) == 0) {
		sprintf(msg, "error: can't open input file '%s'\n", filename);
		return 2;
	} 
	for(block = 0; fread(buffer, 1, sizeof buffer, file) == sizeof buffer; block++) {
		if (block >= num_blocks)
			break;
		if(buffer[0x14] & 0x10) 
			result = 1; // scrambled
		// make hash
		unsigned char* p=buffer;
		for (int i=0; i < sizeof buffer; i++, p++)
			hash += (*p & 0xff);
	}
	fclose(file);
	if (block == 0) { // didn't read anything
		sprintf(msg, "error: can't read from input file '%s'\n"
				"(If reading from a DVD-ROM, make sure the drive is unlocked)", filename);
		return 2;
	}
	return result;
}

// initializes reverse table
void VobInit()
{
	int	loop0, loop1, value; 
	for(loop0 = 0; loop0 != 0x100; loop0++) {
		for(loop1 = value = 0; loop1 != 8; loop1++) {
			value |= ((loop0 >> loop1) & 0x01) << (7 - loop1);
		}
		_reverse[loop0] = (unsigned char) value;
	}
}

// initializes init cache, sets description mode (off if key = NULL, on otherwise)
void VobInitKey(int cache_size, unsigned char* key)
{
	if (key) {
		vrt.decrypt = true;
		vrt.lfsr0 = 0x100;
		vrt.lfsr1 = 0x200000;
		_Salt(vrt.lfsr0, vrt.lfsr1, key);
		if (cache_size <= 0)
			cache_size = 1; // need cache
	}
	else {
		vrt.decrypt = false;
		vrt.lfsr0 = vrt.lfsr1 = 0;
		cache_size = 0; // no need
	}
	if (vrt.buf)
		delete[] vrt.buf;
	if (cache_size > 0) {
		vrt.bufsize = cache_size * 0x800;
		vrt.buf = new unsigned char[vrt.bufsize];
	}
	else {
		vrt.bufsize = 0;
		vrt.buf = 0;
	}
}

// retrieves the (cached) block at pos, stores size
unsigned char* GetVobBlock(FILE* handle, __int64 pos, int& size)
{
	static __int64 last_pos=-1;
	static FILE* last_handle=0;
	static int last_size;

	if ( (pos < 0) || (pos != last_pos) || (handle != last_handle)) {
		if (pos >= 0)
			fsetpos(handle, &pos);
		else
			fgetpos(handle, &pos);
		last_size = fread(vrt.buf, 1, vrt.bufsize, handle);
		if (vrt.decrypt)
			_BufDecrypt(vrt.buf, last_size, vrt.lfsr0, vrt.lfsr1);
		last_handle = handle;
		last_pos = pos;
	}
	size = last_size;
	return vrt.buf;
}
